USE @DBNAME@
GO

DROP PROCEDURE dbo.oo_prune_run_history
GO

CREATE PROCEDURE dbo.oo_prune_run_history 
    @keep_this_many_hours INTEGER = 2160, -- 90 days
    @prune_batch_size INTEGER = 1000,
    @prune_dashboards VARCHAR(5) = 'true',
    @recompute_flow_metrics VARCHAR(5) = 'true',
    @verbose INTEGER = 1
/*
* This procedure attempts to prune the run histories whose start_time
* was prior to a specified date. Please configure the parameters below 
* according to your needs.
* 
* The procedure will recompute the counters in the flow_metrics table.
* 
* The script will execute a series of batch deletes to minimize the
* size of the transaction. The default size of the batch is 1000 histories
* at one time (note the records deleted is a lot more than 1000 as the
* associated run steps and their dependents are also deleted -- in other 
* words, the script deletes 1000 history trees at a time). 
*
* PARAMETERS:
*
* @keep_this_many_hours: sets the number of hours of history retained, relative to the last known start time.
* Histories for runs whose start_time is less than (max(run_history.start_time) - @keep_this_many_hours)
* will be deleted, unless their runs have not completed. 
* Defaults to 2160 (90 days).
*       Example: @keep_this_many_hours = 24 -- keep only the last day.
* 
* @prune_batch_size (default 1000): the batch size for pruning, in terms 
* of number of run histories being deleted.
*       Example: @prune_batch_size = 1000
*
* @prune_dashboards: if set to 'true', the script will also delete data 
* that supports dashboards. Otherwise it will leave the data in place, 
* but because the supporting histories are deleted, links from the dashboards 
* may not always produce the proper reports.
*      Example: @prune_dashboards = 'false'
*
* @recompute_flow_metrics: if set to 'true', after each batch of run histories
* is deleted, the flow_metrics table is update to reflect the new counts.
*      Example: @recompute_flow_metrics = 'false'
* 
* @verbose: if set to a value greater than 0, the script will output messages
* about the progress of deletion, as follows:
*       if @verbose = 1, it outputs basic information at startup
*       if @verbose = 2, it also outputs information about each batch being 
*                        deleted
*       if @verbose >= 3, it also outputs detailed information about flow 
*                        metrics update
* 
* RESULT: how many histories were pruned (an integer).
*/
AS  
    SET NOCOUNT ON;


    /* set the deadlock priority for this session to low such that in case
       we disturb real runs, we're the losing party  */
    SET DEADLOCK_PRIORITY LOW;                  
                                                
    -- validate input params                    
    IF (@prune_batch_size < 2) BEGIN            
        RAISERROR('Invalid pruning batch size, must be at least 2 rows',0,1) WITH NOWAIT;
    END;                                        
    IF (@keep_this_many_hours < 1) BEGIN        
        RAISERROR('Invalid time window, must be at least 1hr',0,1) WITH NOWAIT;
    END;                                        
                                                
    DECLARE @msg VARCHAR(1000),                 
            @prune_start_time DATETIME,         
            @batch_start_time DATETIME,         
            @seconds INTEGER,                   
            @max_start_time DATETIME,           
            @last_start_time DATETIME;          
                                                
    SET @prune_start_time = GETDATE();          
                                                
    /* declare local table variable to store the ids for what we want deleted. */
    DECLARE @oo_pruning_table TABLE (           
        oid integer identity not null primary   key, 
        run_id numeric(19,0),                   
        flow_uuid varchar(255) COLLATE DATABASE_DEFAULT,
        execution_state int,                    
        run_time_millis numeric(19,0));          
                                                
    /* populate the pruning table with increasing id's taken from the 
     * run_history table. Exclude histories whose runs are still active 
     * inside the run table                     
     */                                         
    SELECT @max_start_time = max(start_time) from run_history with (nolock);
                                                
    SET @last_start_time = DATEADD(hour, (-1 *  @keep_this_many_hours), @max_start_time);
                                                
    IF (@verbose > 0)                           
    BEGIN                                       
        SET @msg = 'Preparing pruning table. Will delete histories where start_time <= '
            + CAST (@last_start_time AS VARCHAR);
        RAISERROR(@msg, 0,1) WITH NOWAIT;       
    END;                                      
                                                
    INSERT INTO @oo_pruning_table (run_id, flow_uuid, execution_state, run_time_millis)
     SELECT h.oid, h.flow_uuid, h.execution_state, cast(h.run_time_millis as numeric(19,0))
       FROM dbo.run_history h (nolock)          
      WHERE h.start_time  <= @last_start_time   
        AND NOT EXISTS (SELECT 1 FROM dbo.run r (nolock) WHERE r.history_id = h.oid)
      ORDER BY h.oid ASC;
    
    /* this is how many records we try to prune (total) */
    DECLARE @prune_size INTEGER;
    SELECT @prune_size=COUNT(*) FROM @oo_pruning_table;
    
    IF (@verbose > 0) 
    BEGIN
        SET @msg = 'Pruning set size is: ' + CAST (@prune_size as VARCHAR);
        RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    
    DECLARE @batch_start INTEGER;
    SET @batch_start = 1;
    
    DECLARE @min_id INTEGER;
    DECLARE @max_id INTEGER;
    
    WHILE (@prune_size > 0 AND @batch_start <= @prune_size) 
    BEGIN
            SET @batch_start_time = GETDATE();
            SET @min_id = @batch_start;
            SET @max_id = @batch_start + @prune_batch_size - 1;
    
            IF (@verbose > 1) 
            BEGIN
              SET @msg = 'Deleting chunk: ' + CAST(@min_id AS VARCHAR) + ' to ' + CAST(@max_id AS VARCHAR);
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;

            BEGIN TRANSACTION;
               
               IF (LOWER(@prune_dashboards) = 'true') 
               BEGIN

                  IF (@verbose > 1) RAISERROR('Deleting dashboard data...' , 0,1) WITH NOWAIT;

                  DELETE dbo.log_record
                    FROM dbo.log_record l
                  INNER JOIN @oo_pruning_table p
                     ON ((p.oid BETWEEN @min_id AND @max_id) AND l.run_hist_id = p.run_id)

               END;
    
               IF (@verbose > 1) RAISERROR('Deleting run history...' , 0,1) WITH NOWAIT;
               DELETE dbo.run_history
                 FROM dbo.run_history r
                INNER JOIN @oo_pruning_table p
                   ON ((p.oid BETWEEN @min_id AND @max_id) AND r.oid = p.run_id)

               IF (@verbose > 1) RAISERROR('Updating flow metrics...' , 0,1) WITH NOWAIT;
    
               -- update flow metrics if required:
               IF (LOWER(@recompute_flow_metrics) = 'true')
               BEGIN 

                    IF (@verbose > 2)
                    BEGIN
                      RAISERROR('BEFORE:' , 0,1) WITH NOWAIT;
                      SELECT * from flow_metrics with (nolock);    
                    END;  
                  
                    -- update the flow metrics table to subtract the counts for the run history rows
                    -- that we just deleted.
                    UPDATE dbo.flow_metrics
                      SET diagnosed_count = diagnosed_count - d.diagnosedCount,
                          error_count = error_count - d.errorCount,
                          failed_count = failed_count - d.failedCount,
                          no_action_count = no_action_count - d.noActionCount,
                          resolved_count = resolved_count - d.resolvedCount,
                          cumulative_time = cumulative_time - d.cumulativeTime,
                          dlm_time = GETDATE()
                     FROM dbo.flow_metrics 
                    INNER JOIN (
                     SELECT flow_uuid,
                            sum(case when execution_state = 0 then 1 else 0 end) as diagnosedCount,
                            sum(case when execution_state = 1 then 1 else 0 end) as resolvedCount,
                            sum(case when execution_state = 2 then 1 else 0 end) as noActionCount,
                            sum(case when execution_state = 3 then 1 else 0 end) as errorCount,
                            sum(case when execution_state = 2147483647 then 1 else 0 end) as failedCount,
                            sum(run_time_millis) as cumulativeTime
                       FROM @oo_pruning_table
                      WHERE oid BETWEEN @min_id AND @max_id
                      GROUP BY flow_uuid
                   ) AS d 
                   ON dbo.flow_metrics.flow_uuid = d.flow_uuid AND dbo.flow_metrics.flow_version = 0;
      
                   -- now delete the metrics for those flows that are left with 0 counts across the board
                   DELETE FROM dbo.flow_metrics
                         WHERE diagnosed_count = 0
                           AND failed_count = 0
                           AND no_action_count = 0
                           AND resolved_count = 0
                           AND error_count = 0
                           AND EXISTS (SELECT 1 FROM @oo_pruning_table p
                                        WHERE oid BETWEEN @min_id AND @max_id
                                          AND flow_uuid = p.flow_uuid);

              
                 IF (@verbose > 2) 
                 BEGIN
                   RAISERROR('AFTER:' , 0,1) WITH NOWAIT;
                   SELECT * from flow_metrics with (nolock);    
                 END;

            END; -- update flow metrics
                
            COMMIT TRANSACTION;
    
            SET @batch_start = @batch_start + @prune_batch_size ;   

            IF (@verbose > 1) 
            BEGIN
               SELECT @seconds = DATEDIFF(second, @batch_start_time, GETDATE());
               SET @msg = 'Batch pruning time was: ' + CAST(@seconds as VARCHAR) + ' seconds';
               RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
    END;

    IF (@verbose > 0) 
    BEGIN    
        SELECT @seconds = DATEDIFF(second, @prune_start_time, GETDATE());
        SET @msg = 'Total pruning time was: ' + CAST(@seconds as VARCHAR) + ' seconds';
        RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    RETURN @prune_size

-- END PROCEDURE

GO